@@ -3,7 +3,7 @@
  *
  * Copyright (C) 1990, 1991 John W. Eaton.
  * Copyright (C) 1994, 1995 Graeme W. Wilford. (Wilf.)
- * Copyright (C) 2001, 2002, 2003, 2004, 2006, 2007, 2008 Colin Watson.
+ * Copyright (C) 2001, 2002, 2003, 2004, 2006, 2007, 2008, 2009 Colin Watson.
  *
  * This file is part of man-db.
  *
@@ -59,6 +59,7 @@
 
 #include "canonicalize.h"
 #include "xgetcwd.h"
+#include "xvasprintf.h"
 
 #include "gettext.h"
 #define _(String) gettext (String)
@@ -94,6 +95,7 @@ static char *tmplist[MAXDIRS];
 
 char *user_config_file = NULL;
 int disable_cache;
+int min_cat_width = 80, max_cat_width = 80, cat_width = 0;
 
 static void mkcatdirs (const char *mandir, const char *catdir);
 static inline char *get_manpath_from_path (const char *path);
@@ -347,12 +349,6 @@ char *cat_manpath (char *manp)
 	return catp;
 }		
 
-static char *
-add_to_manpath (char *manpath, const char *path)
-{
-	return pathappend (manpath, path);
-}
-
 /* Unpack a glibc-style locale into its component parts.
  *
  * This function was inspired by _nl_explode_name in libintl; I've rewritten
@@ -427,41 +423,33 @@ void free_locale_bits (struct locale_bits *bits)
 }
 
 
-char *add_nls_manpath (char *manpathlist, const char *locale)
+static char *get_nls_manpath (const char *manpathlist, const char *locale)
 {
-	char *manpath = NULL;
-	char *path;
 	struct locale_bits lbits;
-	char *omanpathlist;
-	char *manpathlist_ptr = manpathlist;
+	char *manpath = NULL;
+	char *manpathlist_copy, *path, *manpathlist_ptr;
 
-	debug ("add_nls_manpath(): processing %s\n", manpathlist);
-
-	if (locale == NULL || *locale == '\0')
-		return manpathlist;
 	unpack_locale_bits (locale, &lbits);
 	if (STREQ (lbits.language, "C") || STREQ (lbits.language, "POSIX")) {
 		free_locale_bits (&lbits);
-		return manpathlist;
+		return xstrdup (manpathlist);
 	}
 
-	omanpathlist = xstrdup (manpathlist);
-
+	manpathlist_copy = xstrdup (manpathlist);
+	manpathlist_ptr = manpathlist_copy;
 	for (path = strsep (&manpathlist_ptr, ":"); path;
 	     path = strsep (&manpathlist_ptr, ":")) {
 		DIR *mandir = opendir (path);
+		struct dirent *mandirent;
+
 		if (!mandir)
 			continue;
 
-		for (;;) {
-			struct dirent *mandirent = readdir (mandir);
+		while ((mandirent = readdir (mandir)) != NULL) {
 			const char *name;
 			struct locale_bits mbits;
 			char *fullpath;
 
-			if (!mandirent)
-				break;
-
 			name = mandirent->d_name;
 			if (STREQ (name, ".") || STREQ (name, ".."))
 				continue;
@@ -479,20 +467,94 @@ char *add_nls_manpath (char *manpathlist, const char *locale)
 			     STREQ (lbits.territory, mbits.territory)) &&
 			    (!*mbits.modifier ||
 			     STREQ (lbits.modifier, mbits.modifier)))
-				manpath = add_to_manpath (manpath, fullpath);
+				manpath = pathappend (manpath, fullpath);
 			free_locale_bits (&mbits);
 			free (fullpath);
 		}
+
+		if (STREQ (lbits.language, "en"))
+			/* For English, we look in the subdirectories as
+			 * above just in case there's something like
+			 * en_GB.UTF-8, but it's more probable that English
+			 * manual pages reside at the top level.
+			 */
+			manpath = pathappend (manpath, path);
 	}
 
 	free_locale_bits (&lbits);
+	return manpath;
+}
 
-	/* After doing all the locale stuff we add the manpath to the *END*
-	 * so the locale dirs are checked first on each section */
-	manpath = add_to_manpath (manpath, omanpathlist);
-	free (omanpathlist);
+char *add_nls_manpaths (char *manpathlist, const char *locales)
+{
+	char *manpath = NULL;
+	char *locales_copy, *tok, *locales_ptr;
+	char *locale_manpath;
+
+	debug ("add_nls_manpaths(): processing %s\n", manpathlist);
+
+	if (locales == NULL || *locales == '\0')
+		return manpathlist;
+
+	/* For each locale, we iterate over the manpath and find appropriate
+	 * locale directories for each item. We then concatenate the results
+	 * for all locales. In other words, LANGUAGE=fr:de and
+	 * manpath=/usr/share/man:/usr/local/share/man could result in
+	 * something like this list:
+	 *
+	 *   /usr/share/man/fr
+	 *   /usr/local/share/man/fr
+	 *   /usr/share/man/de
+	 *   /usr/local/share/man/de
+	 *   /usr/share/man
+	 *   /usr/local/share/man
+	 *
+	 * This assumes that it's more important to have documentation in
+	 * the preferred language than to have documentation for the correct
+	 * object (in the case where there are different versions of a
+	 * program in different hierarchies, for example). It is not
+	 * entirely obvious that this is the right assumption, but on the
+	 * other hand the other choice is not entirely obvious either. We
+	 * tie-break on "we've always done it this way", and people can use
+	 * 'man -a' or whatever in the occasional case where we get it
+	 * wrong.
+	 *
+	 * We go to no special effort to de-duplicate directories here.
+	 * create_pathlist will sort it out later; note that it preserves
+	 * order in that it keeps the first of any duplicate set in its
+	 * original position.
+	 */
+
+	locales_copy = xstrdup (locales);
+	locales_ptr = locales_copy;
+	for (tok = strsep (&locales_ptr, ":"); tok;
+	     tok = strsep (&locales_ptr, ":")) {
+		if (!*tok)	/* ignore empty fields */
+			continue;
+		debug ("checking for locale %s\n", tok);
+
+		locale_manpath = get_nls_manpath (manpathlist, tok);
+		if (locale_manpath) {
+			if (manpath)
+				manpath = appendstr (manpath, ":",
+						     locale_manpath, NULL);
+			else
+				manpath = xstrdup (locale_manpath);
+			free (locale_manpath);
+		}
+	}
+
+	/* Always try untranslated pages as a last resort. */
+	locale_manpath = get_nls_manpath (manpathlist, "C");
+	if (locale_manpath) {
+		if (manpath)
+			manpath = appendstr (manpath, ":",
+					     locale_manpath, NULL);
+		else
+			manpath = xstrdup (locale_manpath);
+		free (locale_manpath);
+	}
 
-	free (manpathlist);
 	return manpath;
 }
 
@@ -753,6 +815,7 @@ static void add_to_dirlist (FILE *config, int user)
 	char *bp;
 	char buf[BUFSIZ];
 	char key[50], cont[512];
+	int val;
 	int c;
 
 	while ((bp = fgets (buf, BUFSIZ, config))) {
@@ -787,6 +850,12 @@ static void add_to_dirlist (FILE *config, int user)
 		else if (sscanf (bp, "SECTIONS %511[^\n]", cont) == 1)
 			/* Since I keep getting it wrong ... */
 			add_sections (cont);
+		else if (sscanf (bp, "MINCATWIDTH %d", &val) == 1)
+			min_cat_width = val;
+		else if (sscanf (bp, "MAXCATWIDTH %d", &val) == 1)
+			max_cat_width = val;
+		else if (sscanf (bp, "CATWIDTH %d", &val) == 1)
+			cat_width = val;
 	 	else {
 			error (0, 0, _("can't parse directory list `%s'"), bp);
 			gripe_reading_mp_config (CONFIG_FILE);
@@ -882,7 +951,8 @@ static char *def_path (int flag)
  * directories listed in the man_db.config file.  If so, and it is
  * not already in the manpath, add it.  If the directory is not listed
  * in the man_db.config file, see if there is a subdirectory `../man' or
- * `man'.  If so, and it is not already in the manpath, add it.
+ * `man', or, for FHS-compliance, `../share/man' or `share/man'.  If so,
+ * and it is not already in the manpath, add it.
  * Example:  user has $HOME/bin in his path and the directory
  * $HOME/man exists -- the directory $HOME/man will be added
  * to the manpath.
@@ -937,13 +1007,15 @@ static inline char *get_manpath_from_path (const char *path)
 
 		 	t = has_mandir (p);
 		 	if (t) {
-				debug ("but does have a ../man or man "
+				debug ("but does have a ../man, man, "
+				       "../share/man, or share/man "
 				       "subdirectory\n");
 	
 				add_dir_to_list (tmplist, t);
 				free (t);
 		 	} else
-				debug ("and doesn't have ../man or man "
+				debug ("and doesn't have ../man, man, "
+				       "../share/man, or share/man "
 				       "subdirectories\n");
 		}
 	}
@@ -1018,34 +1090,41 @@ static void add_dir_to_list (char **lp, const char *dir)
 	}
 }
 
-/* path does not exist in config file: check to see if path/../man or 
-   path/man exist.  If so return it, if not return NULL. */
+/* path does not exist in config file: check to see if path/../man,
+   path/man, path/../share/man, or path/share/man exist.  If so return
+   it, if not return NULL. */
 static inline char *has_mandir (const char *path)
 {
-	char *newpath = NULL;
+	char *newpath;
 
 	/* don't assume anything about path, especially that it ends in 
 	   "bin" or even has a '/' in it! */
 	   
 	char *subdir = strrchr (path, '/');
 	if (subdir) {
-		const int prefix_len = subdir + 1 - path;
-		newpath = xmalloc (prefix_len + sizeof ("man") + 1);
-		strncpy (newpath, path, prefix_len);
-		strcpy (newpath + prefix_len, "man");
-
+		newpath = xasprintf ("%.*s/man", subdir - path, path);
 		if (is_directory (newpath) == 1)
 			return newpath;
-		else
-			*newpath = '\0';
+		free (newpath);
 	}
 
-	newpath = appendstr (newpath, path, "/man", NULL);
-
+	newpath = appendstr (NULL, path, "/man", NULL);
 	if (is_directory (newpath) == 1)
 		return newpath;
-
 	free (newpath);
+
+	if (subdir) {
+		newpath = xasprintf ("%.*s/share/man", subdir - path, path);
+		if (is_directory (newpath) == 1)
+			return newpath;
+		free (newpath);
+	}
+
+	newpath = appendstr (NULL, path, "/share/man", NULL);
+	if (is_directory (newpath) == 1)
+		return newpath;
+	free (newpath);
+
 	return NULL;
 }
 
@@ -1104,7 +1183,7 @@ void create_pathlist (const char *manp, char **mp)
 	/* Eliminate duplicates due to symlinks. */
 	mp = mphead;
 	while (*mp) {
-		char *target;
+		char *target, *oldmp = NULL;
 		char **dupcheck;
 		int found_dup = 0;
 
@@ -1112,15 +1191,21 @@ void create_pathlist (const char *manp, char **mp)
 		 * manpath?
 		 */
 		target = canonicalize_file_name (*mp);
-		if (!target) {
-			++mp;
-			continue;
+		if (target) {
+			oldmp = *mp;
+			*mp = target;
 		}
-		for (dupcheck = mphead; *dupcheck; ++dupcheck) {
-			if (mp == dupcheck || !STREQ (target, *dupcheck))
+		/* Only check up to the current list position, to keep item
+		 * order stable across deduplication.
+		 */
+		for (dupcheck = mphead; *dupcheck && dupcheck != mp;
+		     ++dupcheck) {
+			if (!STREQ (*mp, *dupcheck))
 				continue;
-			debug ("Removing duplicate manpath entry %s -> %s\n",
-			       *mp, target);
+			debug ("Removing duplicate manpath entry %s (%d) -> "
+			       "%s (%d)\n",
+			       oldmp, mp - mphead,
+			       *dupcheck, dupcheck - mphead);
 			free (*mp);
 			for (dupcheck = mp; *(dupcheck + 1); ++dupcheck)
 				*dupcheck = *(dupcheck + 1);
@@ -1128,10 +1213,25 @@ void create_pathlist (const char *manp, char **mp)
 			found_dup = 1;
 			break;
 		}
-		free (target);
+		if (oldmp)
+			free (oldmp);
 		if (!found_dup)
 			++mp;
 	}
+
+	if (debug_level) {
+		int first = 1;
+
+		debug ("final search path = ");
+		for (mp = mphead; *mp; ++mp) {
+			if (first) {
+				debug ("%s", *mp);
+				first = 0;
+			} else
+				debug (":%s", *mp);
+		}
+		debug ("\n");
+	}
 }
 
 void free_pathlist (char **mp)
