@@ -33,7 +33,7 @@ struct locale_bits {
 extern char *cat_manpath (char *manp);
 extern void unpack_locale_bits (const char *locale, struct locale_bits *bits);
 extern void free_locale_bits (struct locale_bits *bits);
-extern char *add_nls_manpath (char *manpathlist, const char *locale);
+extern char *add_nls_manpaths (char *manpathlist, const char *locales);
 extern char *get_manpath (const char *systems);
 extern void create_pathlist (const char *manp, char **mp);
 extern void free_pathlist (char **mp);
